import tqdm
import time
import threading

__all__ = ["Simulation"]


class Simulation:
    '''
    Simulation class for easily creating and running agent-based simulations.
    Includes progress monitors for terminal and Jupyter notebooks.
    '''

    # noinspection PyUnresolvedReferences
    def __init__(self, *args):
        '''
        Initialize the simulation

        :param args: unpacked list of agents, e.g. Simulation(alice, bob, charlie). All agents must share the same
                     output dictionary using Agent.shared_output()
        '''
        self.out = args[0].out
        self.agents = args
        try:  # figure out if we're in a Jupyter notebook or not
            __IPYTHON__
            self.is_ipython = True
        except:
            self.is_ipython = False

    def progress_monitor(self, poison_pill):
        '''
        Display a tqdm-style progress bar in a Jupyter notebook

        :param threading.Event poison_pill: a flag to kill the progressMonitor thread
        '''
        pbars = {}
        progress = {}
        progress_max = {}
        for agent in self.agents:
            if self.is_ipython:
                pbars[agent.name] = tqdm.tqdm_notebook(total = len(agent.stream), desc = agent.name)
            else:
                pbars[agent.name] = tqdm.tqdm(total = len(agent.stream), desc = agent.name)
            progress[agent.name] = 0
            progress_max[agent.name] = len(agent.stream)

        # Loop and update progress
        while not poison_pill.isSet():
            for agent in self.agents:
                dProg = self.out[agent.name + ":progress"] - progress[agent.name]
                progress[agent.name] += dProg
                pbars[agent.name].update(dProg)
            time.sleep(0.05)

        for agent in self.agents:
            pbars[agent.name].n = pbars[agent.name].total
            pbars[agent.name].close()

    # noinspection PyUnboundLocalVariable
    def run(self, monitor_progress = True):
        '''
        Run the simulation

        :param monitor_progress: whether to display a progress bar for each agent
        '''

        for agent in self.agents:
            agent.start()

        if monitor_progress:
            poison_pill = threading.Event()
            progress_monitor = threading.Thread(target = self.progress_monitor, args = (poison_pill,))
            progress_monitor.start()

        for agent in self.agents:
            agent.join()

        if monitor_progress:
            poison_pill.set()
            progress_monitor.join()



import sys
import numpy as np
import multiprocessing
import ctypes
from multiprocessing import sharedctypes
from squanch import channels, qstream

__all__ = ["Agent"]


class Agent(multiprocessing.Process):
    '''
    Represents an entity (Alice, Bob, etc.) that can send messages over classical and quantum communication channels.
    Agents have the following properties:

    * Incoming and outgoing classical communciation lines to other agents
    * Incoming and outgoing quantum channels to other agents through which entangled pairs may be distributed
    * Ideal classical memory
    * Quantum memory with some characteristic corruption timescale
    '''

    def __init__(self, hilbert_space, out = None, name = None, data = None):
        '''
        Instantiate an Agent from a unique identifier and a shared memory pool

        :param np.array hilbert_space: the shared memory pool representing the Hilbert space of the qstream
        :param dict out: shared output dictionary to pass to Agent processes to allow for "returns". Default: {}
        :param str name: the unique identifier for the Agent. Default: class name
        :param any data: data to pass to the Agent's process, stored in ``self.data``. Default: None
        '''
        multiprocessing.Process.__init__(self)
        # Name of the agent, e.g. "Alice". Defaults to the name of the class.
        if name is not None:
            self.name = name
        else:
            self.name = self.__class__.__name__

        # Agent's clock
        self.time = 0.0
        # self.retardedTime = 0.0
        self.pulse_length = 10 * 10 ** -12  # 10ps photon pulse size

        # Register input data and output structure
        self.data = data
        if out is None:
            out = {}
        out[self.name] = None
        out[self.name + ":progress"] = 0
        out[self.name + ":progress_max"] = hilbert_space.shape[0]
        self.stream = qstream.QStream.from_array(hilbert_space, agent = self)
        self.out = out

        # Communication channels are dicts; keys: agent objects, values: channel objects
        self.cchannels_in = {}
        self.cchannels_out = {}
        self.qchannels_in = {}
        self.qchannels_out = {}

        # Classical memory is a large 1D array
        # classicalMemorySize = 2 ** 16
        self.cmem = {}  # np.zeros(classicalMemorySize)

        # Quantum memory is an array of "blocks". Each element in a block holds a qubit
        # qBlockSize = 256
        # numQBlocks = 64
        self.qmem = {}  # ((None,) * qBlockSize,) * numQBlocks
        # self.qDecayTimescale = 100.0  # Coherence timescale for qubits in quantum memory

    def __hash__(self):
        '''
        Agents are hashed by their names, which is why they must be unique.
        '''
        return hash(self.name)

    def __eq__(self, other):
        '''
        Agents are compared for equality by their names.
        '''
        return self.name == other.name

    def __ne__(self, other):
        '''
        Overridden inequality operator, for good practice.
        '''
        return not (self == other)

    @staticmethod
    def shared_output():
        '''
        Generate a shared output dictionary to distribute among agents in separate processes

        :return: an empty multiprocessed Manager.dict()
        '''
        return multiprocessing.Manager().dict()

    @staticmethod
    def shared_hilbert_space(system_size, num_systems):
        '''
        Allocate a portion of shareable c-type memory to create a numpy array that is sharable between processes

        :param int system_size: number of entangled qubits in each quantum system; each has dimension 2^system_size
        :param int num_systems: number of small quantum systems in the data stream
        :return: a blank, sharable, num_systems * 2^system_size * 2^system_size array of np.complex64 values
        '''
        dim = 2 ** system_size
        mallocced = sharedctypes.RawArray(ctypes.c_double, num_systems * dim * dim)
        array = np.frombuffer(mallocced, dtype = np.complex64).reshape((num_systems, dim, dim))
        qstream.QStream.reformat(array)
        return array

    def qconnect(self, other, channel = channels.QChannel, **kwargs):
        '''
        Connect Alice and Bob bidirectionally via a simulated quantum channel

        :param Agent other: the other agent to connect to
        :param QChannel channel: the quantum channel model to use
        :param \**kwargs: optional channel arguments
        '''
        # Instantiate quantum channels between Alice and Bob
        qchannel_alice_to_bob = channel(self, other, **kwargs)
        qchannel_bob_to_alice = channel(other, self, **kwargs)
        self.qchannels_out[other] = qchannel_alice_to_bob
        self.qchannels_in[other] = qchannel_bob_to_alice
        other.qchannels_out[self] = qchannel_bob_to_alice
        other.qchannels_in[self] = qchannel_alice_to_bob
        # Make a section of Alice's/Bob's quantum memory for Bob/Alice
        self.qmem[other] = []
        other.qmem[self] = []

    def qsend(self, target, qubit):
        '''
        Send a qubit to another agent. The qubit is serialized and passed through a QChannel to the
        targeted agent, which can retrieve the qubit with Agent.qrecv(). ``self.time`` is updated upon
        calling this method.

        :param Agent target: the agent to send the qubit to
        :param Qubit qubit: the qubit to send
        '''
        self.qchannels_out[target].put(qubit)
        self.time += self.pulse_length

    def qrecv(self, origin):
        '''
        Receive a qubit from another connected agent. ``self.time`` is updated upon calling this method.

        :param Agent origin: The agent that previously sent the qubit
        :return: the retrieved qubit, which is also stored in ``self.qmem``
        '''
        qubit, recvTime = self.qchannels_in[origin].get()
        # Update agent clock
        self.time = max(self.time, recvTime)
        # Add qubit to quantum memory
        self.qmem[origin].append(qubit)
        return qubit

    def qstore(self, qubit):
        '''
        Store a qubit in quantum memory. Equivalent to ``self.qmem[self].append(qubit)``.

        :param Qubit qubit: the qubit to store
        '''
        self.qmem[self].append(qubit)

    def cconnect(self, other, channel = channels.CChannel, **kwargs):
        '''
        Connect Alice and Bob bidirectionally via a simulated classical channel

        :param Agent other: the other agent to connect to
        :param CChannel channel: the classical channel model to use
        :param \**kwargs: optional channel arguments
        '''
        # Instantiate classical channels between Alice and Bob
        cchannel_alice_to_bob = channel(self, other, **kwargs)
        cchannel_bob_to_alice = channel(other, self, **kwargs)
        self.cchannels_out[other] = cchannel_alice_to_bob
        self.cchannels_in[other] = cchannel_bob_to_alice
        other.cchannels_out[self] = cchannel_bob_to_alice
        other.cchannels_in[self] = cchannel_alice_to_bob
        # Make a section of Alice's/Bob's classical memory for Bob/Alice
        self.cmem[other] = []
        other.cmem[self] = []

    def csend(self, target, thing):
        '''
        Send a serializable object to another agent. The transmission time is updated by (number of bits) pulse lengths.

        :param Agent target: the agent to send the transmission to
        :param any thing: the object to send
        '''

        self.cchannels_out[target].put(thing)
        self.time += sys.getsizeof(thing) * 8 * self.pulse_length

    def crecv(self, origin):
        '''
        Receive a serializable object from another connected agent. ``self.time`` is updated upon calling this method.

        :param Agent origin: The agent that previously sent the qubit
        :return: the retrieved object, which is also stored in ``self.cmem``
        '''
        thing, recvTime = self.cchannels_in[origin].get()
        # Update agent clock
        self.time = max(self.time, recvTime)
        # Add qubit to quantum memory
        self.cmem[origin].append(thing)
        return thing

    def run(self):
        '''This method should be overridden in extended class instances, and cannot take any arguments or
        return any values.'''
        pass

    def output(self, thing):
        '''
        Output something to ``self.out[self.name]``

        :param any thing: the thing to put in the dictionary
        '''
        self.out[self.name] = thing

    def update_progress(self, value):
        '''
        Update the progress of this agent in the shared output dictionary. Used in Simulation.progress_monitor().

        :param value: the value to update the progress to (out of a max of len(self.stream))
        '''
        self.out[self.name + ":progress"] = value

    def increment_progress(self):
        '''
        Adds 1 to the current progress
        '''
        self.out[self.name + ":progress"] += 1



import numpy as np
from squanch import qubit, linalg

__all__ = ["QStream"]


def zero_state(system_size, num_systems):
    '''
    Generate an array representing the num_systems Hilbert spaces in the state ``|0>...|0><0|...<0|``

    :param int system_size: maximum size of entangled subsystems
    :param int num_systems: number of disjoint quantum subsystems to allocate
    :return: the all-zero state array
    '''
    # Initialize each qubit state
    zero = np.array([1, 0], dtype = np.complex64)  # vector representation of the |0> pure state
    initial_qubit_state = np.outer(zero, zero)  # each qubit is initialized as |0><0|
    initial_system_state = np.array([], dtype = np.complex64)
    # Generate the matrix representation of the initial state of the n-qubit system
    for _ in range(system_size):
        initial_system_state = linalg.tensor_product(initial_system_state, initial_qubit_state)
    # The Hilbert space of QSys0 x QSys1 x ... x QSysN is represented as an n-dimensional array of state vectors
    return np.repeat([initial_system_state], num_systems, axis = 0)  # repeat state into vertical axis


class QStream:
    '''
    Efficiently handle a large number of small entangled quantum systems to avoid having to perform many class
    instantiations when simulating transmission of data through quantum channels
    '''

    def __init__(self, system_size, num_systems, array = None, agent = None):
        '''
        Instantiate the quantum datastream object

        :param int system_size: number of entangled qubits in each quantum system; each system has dims 2^system_size
        :param int num_systems: number of small quantum systems in the data stream
        :param np.array array: pre-allocated array in memory for purposes of sharing QStreams in multiprocessing
        :param Agent agent: optional reference to the Agent owning the qstream; useful for progress monitoring across
                            separate processes
        '''
        self.system_size = system_size  # number of qubits per system
        self.num_systems = num_systems  # number of disjoint quantum subsystems
        self.agent = agent
        # Generate the matrix representation of the overall state of the quantum stream
        if array is not None:
            self.state = array
        else:
            self.state = zero_state(system_size, num_systems)

        # The "head" of the stream; what qsystem is being processed at the moment
        self.index = 0

    def __iter__(self):
        '''
        Custom iterator method for streams

        :return: each system in the stream
        '''
        for i in range(self.num_systems):
            if self.agent: self.agent.update_progress(i)
            yield self.system(i)

    def __len__(self):
        '''
        Custom length method for streams; equivalent to stream.num_systems

        :return: stream.num_systems
        '''
        return self.num_systems

    @classmethod
    def from_array(cls, array, reformat = False, agent = None):
        '''
        Instantiates a quantum datastream object from a (typically shared) pre-allocated array

        :param np.array array: the pre-allocated np.complex64 array representing the shared Hilbert space
        :param bool reformat: if providing a pre-allocated array, whether to reformat it to the all-zero state
        :return: the child QStream
        '''
        num_systems = array.shape[0]
        system_size = int(np.log2(array.shape[1]))
        qstream = cls(system_size, num_systems, array = array, agent = agent)
        if reformat:
            qstream.reformat(qstream.state)
        return qstream

    @staticmethod
    def reformat(array):
        '''
        Reformats a Hilbert space array in-place to the all-zero state

        :param np.array array: a num_systems x 2^system_size x 2^system_size array of np.complex64 values
        '''
        num_systems = array.shape[0]
        system_size = int(np.log2(array.shape[1]))
        array[...] = zero_state(system_size, num_systems)

    def system(self, index):
        '''
        Access the nth quantum system in the quantum datastream object

        :param int index: zero-index of the quantum system to access
        :return: the quantum system
        '''
        return qubit.QSystem.from_stream(self, index)

    def next(self):
        '''
        Access the next element in the quantum stream, returning it as a QSystem object, and increment the head by 1

        :return: a QSystem for the "head" system
        '''
        sys = self.system(self.index)
        self.index += 1
        return sys



import sys
import multiprocessing
from squanch.qubit import Qubit
from squanch import errors

__all__ = ["QChannel", "CChannel", "FiberOpticQChannel"]


class QChannel:
    '''
    Base class for a quantum channel
    '''

    def __init__(self, from_agent, to_agent, length = 0.0, errors = ()):
        '''
        Instantiate the quantum channel

        :param Agent from_agent: sending agent
        :param Agent to_agent: receiving agent
        :param float length: length of quantum channel in km; default: 0.0km
        :param QError[] errors: list of error models to apply to qubits in this channel; default: [] (no errors)
        '''
        # Register agent connections
        self.from_agent = from_agent
        self.to_agent = to_agent

        # Signal propagation properties
        self.length = length  # Physical length of the channel in km
        self.signal_speed = 2.998 * 10 ** 5  # Speed of light in km/s

        # A queue representing the qubits in transit along the channel
        self.queue = multiprocessing.Queue()

        # Register error models
        self.errors = errors

    def put(self, qubit):
        '''
        Serialize and push qubit into the channel queue

        :param Qubit qubit: the qubit to send
        '''
        # Calculate the time of arrival
        time_of_arrival = self.from_agent.time + self.from_agent.pulse_length + (self.length / self.signal_speed)
        if qubit is not None:
            self.queue.put((qubit.serialize(), time_of_arrival))
        else:
            self.queue.put((None, time_of_arrival))

    def get(self):
        '''
        Retrieve a qubit by reference from the channel queue, applying errors upon retrieval

        :return: tuple: (the qubit with errors applied (possibly ``None``), receival time)
        '''
        indices, receive_time = self.queue.get()
        if indices is not None:
            system_index, qubit_index = indices
            qubit = Qubit.from_stream(self.to_agent.stream, system_index, qubit_index)
        else:
            qubit = None

        # Apply errors
        for error in self.errors:
            qubit = error.apply(qubit)

        # Return modified qubit and return time
        return qubit, receive_time


class CChannel:
    '''
    Base class for a classical channel
    '''

    def __init__(self, from_agent, to_agent, length = 0.0):
        '''
        Instantiate the quantum channel

        :param Agent from_agent: sending agent
        :param Agent to_agent: receiving agent
        :param float length: length of fiber optic line in km; default: 0.0km
        '''
        # Register agent connections
        self.from_agent = from_agent
        self.to_agent = to_agent

        # Signal propagation
        self.length = length  # Physical length of the channel in km
        self.signal_speed = 2.998 * 10 ** 5  # Speed of light in km/s

        # The channel queue
        self.queue = multiprocessing.Queue()

    def put(self, thing):
        '''
        Serialize and push a serializable object into the channel queue

        :param any thing: the qubit to send
        '''
        # Calculate the time of arrival
        pulse_time = sys.getsizeof(thing) * 8 * self.from_agent.pulse_length
        time_of_arrival = self.from_agent.time + pulse_time + (self.length / self.signal_speed)
        self.queue.put((thing, time_of_arrival))

    def get(self):
        '''
        Retrieve a classical object form the queue

        :return: tuple: (the object, receival time)
        '''
        thing, receive_time = self.queue.get()
        return thing, receive_time


class FiberOpticQChannel(QChannel):
    '''
    Represents a fiber optic line with attenuation errors
    '''

    def __init__(self, from_agent, to_agent, length = 0.0):
        '''
        Instantiate the simulated fiber optic quantum channel

        :param Agent from_agent: sending agent
        :param Agent to_agent: receiving agent
        :param float length: length of fiber optic channel in km; default: 0.0km
        '''
        QChannel.__init__(self, from_agent, to_agent, length = length)

        # Register attenuation errors
        self.errors = [
            errors.AttenuationError(self),
            # errors.RandomUnitaryError(self, 2*np.pi / 100),
            # errors.SystematicUnitaryError(self, 2*np.pi / 10),
        ]



import numpy as np
from squanch import linalg, gates

__all__ = ["QSystem", "Qubit"]

# Computational basis and projection operators
_0 = np.array([1, 0], dtype = np.complex64)
_1 = np.array([0, 1], dtype = np.complex64)
_M0 = np.outer(_0, _0)
_M1 = np.outer(_1, _1)


class QSystem:
    '''
    Represents a multi-particle Hilbert space for several qubuts comprising a single system of a quantum datastream.
    Designed to have similar syntax to QubitSytem, but instantiation is much faster
    '''

    def __init__(self, num_qubits, index = None, state = None):
        '''
        Instatiate the quantum state for an n-qubit system

        :param int num_qubits: number of qubits in the system, treated as maximally entangled
        :param np.array state: density matrix representing the quantum state. By default, |000...0><0...000| is used
        '''
        self.num_qubits = num_qubits
        self.qubits = (Qubit(self, i) for i in range(num_qubits))  # this is a generator, not a list
        self.index = index
        # Register the state or generate a new one
        if state is not None:
            self.state = state  # density matrix should be passed by reference and will modify the QStream.state
        else:
            # Initialize each qubit state
            initial_qubit_state = np.outer(_0, _0)  # each qubit is initialized as |0><0|
            initial_system_state = np.array([], dtype = np.complex64)
            # Generate the matrix representation of the initial state of the n-qubit system
            for _ in range(self.num_qubits):
                initial_system_state = linalg.tensor_product(initial_system_state, initial_qubit_state)
            # Assign the state
            self.state = initial_system_state

    @classmethod
    def from_stream(cls, qstream, index):
        '''
        Instantiate a QSystem from a given point in a parent QStream

        :param QStream qstream: the parent stream
        :param int index: the index in the parent stream corresponding to this system
        :return: the QSystem object
        '''
        return cls(qstream.system_size, index = index, state = qstream.state[index])

    def qubit(self, index):
        '''
        Access a qubit by index; self.qubits does not instantiate all qubits unless casted to a list. Use this
        function to access a single qubit of a given index.

        :param int index: qubit index to generate a qubit instance for
        :return: the qubit instance
        '''
        return Qubit(self, index)

    def measure_qubit(self, index):
        '''
        Measure the qubit at a given index, partially collapsing the state based on the observed qubit value.
        The state vector is modified in-place by this function.

        :param int index: the qubit to measure
        :return: the measured qubit value
        '''
        measure0 = gates.expand(_M0, index, self.num_qubits, "0" + str(index) + str(self.num_qubits))
        prob0 = np.trace(np.dot(measure0, self.state))
        # Determine if qubit collapses to |0> or |1>
        if np.random.rand() <= prob0:
            # qubit collapses to |0>
            self.state[...] = np.linalg.multi_dot([measure0, self.state, measure0.conj().T]) / prob0
            return 0
        else:
            # qubit collapses to |1>
            measure1 = gates.expand(_M1, index, self.num_qubits, "1" + str(self.num_qubits))
            self.state[...] = np.linalg.multi_dot([measure1, self.state, measure1]) / (1.0 - prob0)
            return 1

    def apply(self, operator):
        '''
        Apply an n-qubit operator to this system's n-qubit quantum state: U|psi><psi|U^+ (U^+ = U for Hermitian)

        :param np.array operator: the *Hermitian* n-qubit operator to apply
        :return: nothing, the qsystem state is mutated
        '''
        # Apply the operator
        # assert linalg.isHermitian(operator), "Qubit operators must be Hermitian"
        self.state[...] = np.linalg.multi_dot([operator, self.state, operator.conj().T])
        # self.state[...] = np.linalg.multi_dot([operator, self.state, operator])


class Qubit:
    '''
    Represents a single physical qubit, which is a wrapper for part of a pre-allocated nonlocal QSystem
    '''

    def __init__(self, qsystem, index):
        '''
        Instantiate the qubit from an existing QSystem and index

        :param QSystem qsystem: n-qubit quantum system that this qubit points to
        :param int index: particle index in the quantum system, ranging from 0 to n-1
        '''
        self.index = index
        self.qsystem = qsystem

    @classmethod
    def from_stream(cls, qstream, system_index, qubit_index):
        '''
        Instantiate a qubit from a parent stream (via a QSystem call)

        :param QStream qstream: the parent stream
        :param int system_index: the index corresponding to the parent QSystem
        :param int qubit_index: the index of the qubit to be recalled
        :return: the qubit
        '''
        return qstream.system(system_index).qubit(qubit_index)

    def measure(self):
        '''
        Measure a qubit, mutating its state in-place

        :return: the measured value
        '''
        return self.qsystem.measure_qubit(self.index)

    def getState(self):
        '''
        Traces over the remaining portions of the qsystem to return this qubit's state expressed as a density matrix.

        :return: The (mixed) density matrix describing this qubit's state
        '''
        # TODO: partial trace in numpy?
        pass

    def apply(self, operator, cacheID = None):
        '''
        Apply a single-qubit operator to this qubit, tensoring with I and passing to the qsystem.apply() method

        :param np.array operator: a single qubit (2x2) complex-valued matrix
        :param str cacheID: a character or string to cache the expanded operator by (e.g. Hadamard qubit 2 -> "IHII...")
        '''
        self.qsystem.apply(gates.expand(operator, self.index, self.qsystem.num_qubits, cacheID))

    def serialize(self):
        '''
        Generate a reference to reconstruct this qubit from shared memory

        :return: qubit reference as (systemIndex, qubitIndex)
        '''
        return self.qsystem.index, self.index



import numpy as np

__all__ = ["is_hermitian", "tensor_product", "tensors", "tensor_fill_identity"]


def is_hermitian(matrix):
    '''
    Checks if an operator is Hermitian

    :param np.array matrix: the operator to check
    :return: true or false
    '''
    return np.array_equal(matrix.conj().T, matrix)


def tensor_product(state1, state2):
    '''
    Returns the Kronecker product of two states

    :param np.array state1: the first state
    :param np.array state2: the second state
    :return: the tensor product
    '''
    if len(state1) == 0:
        return state2
    elif len(state2) == 0:
        return state1
    else:
        return np.kron(state1, state2)


def tensors(operator_list):
    '''
    Returns the iterated Kronecker product of a list of states

    :param [np.array] operator_list: list of states to tensor-product
    :return: the tensor product
    '''
    result = []
    for operator in operator_list:
        result = tensor_product(result, operator)
    return result


def tensor_fill_identity(single_qubit_operator, n_qubits, qubit_index):
    '''
    Create the n-qubit operator I x I x ... Operator x I x I... with operator applied to a given qubit index

    :param np.array single_qubit_operator: the operator in the computational basis (a 2x2 matrix)
    :param int n_qubits: the number of qubits in the system to fill
    :param int qubit_index: the zero-indexed qubit to apply this operator to
    :return: the n-qubit operator
    '''
    # assert 0 <= qubit_index <= n_qubits - 1, "qubit Index is out of range"
    operator = tensors([
        tensors([np.eye(2)] * qubit_index),
        single_qubit_operator,
        tensors([np.eye(2)] * (n_qubits - (qubit_index + 1)))
    ])
    return operator



import numpy as np
from squanch import linalg

__all__ = ["H", "X", "Y", "Z", "RX", "RY", "RZ", "CNOT", "TOFFOLI", "CU", "SWAP", "expand"]

# Single qubit operators that can be applied with qubit.apply()

# Projection operators
_M0 = np.outer([1, 0],
               [1, 0])
_M1 = np.outer([0, 1],
               [0, 1])

# Identity operator
_I = np.array([[1, 0],
               [0, 1]])

# Hadamard gate
_H = 1 / np.sqrt(2) * np.array([[1, 1],
                                [1, -1]])

# Pauli-X gate (bit flip)
_X = np.array([[0, 1],
               [1, 0]])

# Pauli-Y gate (bit + phase flip)
_Y = np.array([[0, -1j],
               [1j, 0]])

# Pauli-Z gate (phase flip)
_Z = np.array([[1, 0],
               [0, -1]])


# Single qubit gates
def H(qubit):
    '''
    Applies the Hadamard transform to the specified qubit, updating the qsystem state.
    cache_id: ``H``

    :param Qubit qubit: the qubit to apply the operator to
    '''
    qubit.qsystem.apply(expand(_H, qubit.index, qubit.qsystem.num_qubits, "H"))


def X(qubit):
    '''
    Applies the Pauli-X (NOT) operation to the specified qubit, updating the qsystem state.
    cache_id: ``X``

    :param Qubit qubit: the qubit to apply the operator to
    '''
    qubit.qsystem.apply(expand(_X, qubit.index, qubit.qsystem.num_qubits, "X"))


def Y(qubit):
    '''
    Applies the Pauli-Y operation to the specified qubit, updating the qsystem state.
    cache_id: ``Y``

    :param Qubit qubit: the qubit to apply the operator to
    '''
    qubit.qsystem.apply(expand(_Y, qubit.index, qubit.qsystem.num_qubits, "Y"))


def Z(qubit):
    '''
    Applies the Pauli-Z operation to the specified qubit, updating the qsystem state.
    cache_id: ``Z``

    :param Qubit qubit: the qubit to apply the operator to
    '''
    qubit.qsystem.apply(expand(_Z, qubit.index, qubit.qsystem.num_qubits, "Z"))


def RX(qubit, angle):
    '''
    Applies the single qubit X-rotation operator to the specified qubit, updating the qsystem state.
    cache_id: ``Rx*``, where * is angle/pi

    :param Qubit qubit: the qubit to apply the operator to
    :param float angle: the angle by which to rotate
    '''
    gate = np.cos(angle / 2.0) * _I - 1j * np.sin(angle / 2.0) * _X
    qubit.qsystem.apply(expand(gate, qubit.index, qubit.qsystem.num_qubits, "Rx" + str(angle / np.pi)))


def RY(qubit, angle):
    '''
    Applies the single qubit Y-rotation operator to the specified qubit, updating the qsystem state.
    cache_id: ``Ry*``, where * is angle/pi

    :param Qubit qubit: the qubit to apply the operator to
    :param float angle: the angle by which to rotate
    '''
    gate = np.cos(angle / 2.0) * _I - 1j * np.sin(angle / 2.0) * _Y
    qubit.qsystem.apply(expand(gate, qubit.index, qubit.qsystem.num_qubits, "Ry" + str(angle / np.pi)))


def RZ(qubit, angle):
    '''
    Applies the single qubit Z-rotation operator to the specified qubit, updating the qsystem state.
    cache_id: ``Rz*``, where * is angle/pi

    :param Qubit qubit: the qubit to apply the operator to
    :param float angle: the angle by which to rotate
    '''
    gate = np.cos(angle / 2.0) * _I - 1j * np.sin(angle / 2.0) * _Z
    qubit.qsystem.apply(expand(gate, qubit.index, qubit.qsystem.num_qubits, "Rz" + str(angle / np.pi)))


def CNOT(control, target):
    '''
    Applies the controlled-NOT operation from control on target. This gate takes two qubit arguments to
    construct an arbitrary CNOT matrix.
    cache_id: ``CNOTi,j``, where i and j are control and target indices

    :param Qubit control: the control qubit
    :param Qubit target: the target qubit, with Pauli-X applied according to the control qubit
    '''
    key = "CNOT" + str(control.index) + "," + str(target.index)
    # Generate the gate if needed
    if key not in _expandedGateCache:
        # Represent CNOT(i,j) as |0i><0i| x I x ... x I + |1i><1i| x I x ... x Xtarg x I x ...
        proj0 = linalg.tensor_fill_identity(_M0, control.qsystem.num_qubits, control.index)
        proj1gates = [_I for _ in range(control.qsystem.num_qubits)]
        proj1gates[control.index] = _M1
        proj1gates[target.index] = _X
        proj1 = linalg.tensors(proj1gates)
        CNOTij = proj0 + proj1
        # Cache the gate
        _expandedGateCache[key] = CNOTij
    # Apply the gate
    target.qsystem.apply(_expandedGateCache[key])


def CU(control, target, unitary):
    '''
    Applies the controlled-unitary operation from control on target. This gate takes control and target qubit arguments
    and a unitary operator to apply
    cache_id: ``CUi,j,<str(unitary)>``, where i and j are control and target indices

    :param Qubit control: the control qubit
    :param Qubit target: the target qubit
    :param np.array unitary: the unitary single-qubit gate to apply to the target qubit
    '''
    key = "CU" + str(control.index) + "," + str(target.index) + "," + str(unitary)
    # Generate the gate if needed
    if key not in _expandedGateCache:
        # Represent CNOT(i,j) as |0i><0i| x I x ... x I + |1i><1i| x I x ... x Xtarg x I x ...
        proj0 = linalg.tensor_fill_identity(_M0, control.qsystem.num_qubits, control.index)
        proj1gates = [_I for _ in range(control.qsystem.num_qubits)]
        proj1gates[control.index] = _M1
        proj1gates[target.index] = unitary
        proj1 = linalg.tensors(proj1gates)
        CUij = proj0 + proj1
        # Cache the gate
        _expandedGateCache[key] = CUij
    # Apply the gate
    target.qsystem.apply(_expandedGateCache[key])


def TOFFOLI(control1, control2, target):
    '''
    Applies the Toffoli (or controlled-controlled-NOT) operation from control on target. This gate takes three qubit
    arguments to construct an arbitrary CCNOT matrix.
    cache_id: ``CCNOTi,j,k``, where i and j are control indices and k target indices

    :param Qubit control1: the first control qubit
    :param Qubit control2: the second control qubit
    :param Qubit target: the target qubit, with Pauli-X applied according to the control qubit
    '''
    c1, c2 = sorted([control1.index, control2.index])
    num_qubits = target.qsystem.num_qubits
    key = "CCNOT" + str(c1) + "," + str(c2) + "," + str(target.index)
    # Generate the gate if needed
    if key not in _expandedGateCache:
        gates_list = [[_M0, _M0, _I], [_M0, _M1, _I], [_M1, _M0, _I], [_M1, _M1, _X]]
        CNOTijk = 0
        for gates in gates_list:
            ops = [_I for _ in range(num_qubits)]
            ops[c1] = gates[0]
            ops[c2] = gates[1]
            ops[target.index] = gates[2]
            operator = linalg.tensors(ops)
            CNOTijk += operator
        # Cache the gate
        _expandedGateCache[key] = CNOTijk
    # Apply the gate
    target.qsystem.apply(_expandedGateCache[key])


def SWAP(q1, q2):
    '''
    Applies the SWAP operator to two qubits, switching the states. This gate is implemented by three CNOT operations 
    and thus has no cache_id.
    :param q1: the first qubit
    :param q2: the second qubit
    '''
    CNOT(q2, q1)
    CNOT(q1, q2)
    CNOT(q2, q1)


_expandedGateCache = {}


def expand(operator, index, num_qubits, cache_id = None):
    '''
    Apply a k-qubit quantum gate to act on n-qubits by filling the rest of the spaces with identity operators

    :param np.array operator: the single- or n-qubit operator to apply
    :param int index: if specified, the index of the qubit to perform the operation on
    :param int num_qubits: the number of qubits in the system
    :param str cache_id: a character identifier to cache gates and their expansions in memory
    :return: the expanded n-qubit operator
    '''

    if cache_id is not None:
        key = "I" * index + cache_id + "I" * (num_qubits - 1 - index)
        if key not in _expandedGateCache:  # cache the expanded gate
            expandedOperator = linalg.tensor_fill_identity(operator, num_qubits, index)
            _expandedGateCache[key] = expandedOperator
        return _expandedGateCache[key]
    else:
        return linalg.tensor_fill_identity(operator, num_qubits, index)



import numpy as np
from squanch import gates

__all__ = ["QError", "AttenuationError", "RandomUnitaryError", "SystematicUnitaryError"]


class QError:
    '''A generalized quantum error model'''

    def __init__(self, qchannel):
        '''
        Base initialization class; extend in child methods by overwriting along with QError.__init__(self, qchannel)

        :param qchannel: the quantum channel this error model is being used on
        '''
        self.qchannel = qchannel

    def apply(self, qubit):
        '''
        Generic apply method; overwrite in child methods while maintaining the Qubit->(Qubit | None) signature

        :param Qubit qubit: the qubit being withdrawn from the quantum channel with channel.get(); possibly None
        :return: the modified qubit
        '''
        if qubit is not None:
            # Implement error logic here
            pass
        return qubit


class AttenuationError(QError):
    '''Simulate the possible loss of a qubit in a fiber optic channel due to attenuation effects'''

    def __init__(self, qchannel, attenuation_coefficient = -0.16):
        '''
        Instatiate the error class

        :param QChannel qchannel: parent quantum channel
        :param float attenuation_coefficient: attenuation of fiber in dB/km; default: -.16 dB/km, from Yin, et al
        '''
        QError.__init__(self, qchannel)
        decibel_loss = qchannel.length * attenuation_coefficient
        self.attenuation = 10 ** (decibel_loss / 10)  # Total attenuation along the fiber

    def apply(self, qubit):
        '''
        Simulates possible loss + measurement of qubit

        :param Qubit qubit: qubit from quantum channel
        :return: either unchanged qubit or None
        '''
        if np.random.rand() > self.attenuation and qubit is not None:
            # Photon was lost due to attenuation effects; collapse state and return nothing
            qubit.measure()
            qubit = None
        return qubit


class RandomUnitaryError(QError):
    '''Simualates a random rotation along X and Z with a Gaussian distribution of rotation angles'''

    def __init__(self, qchannel, variance):
        '''
        Instatiate the error class

        :param QChannel qchannel: parent quantum channel
        :param float variance: variance to use in the Gaussian sampling of X and Z rotation angles
        '''
        QError.__init__(self, qchannel)
        self.variance = variance

    def apply(self, qubit):
        '''
        Simulates random rotations on X and Z of a qubit

        :param Qubit qubit: qubit from quantum channel
        :return: rotated qubit
        '''
        if qubit is not None:
            x_angle, z_angle = np.random.normal(0, self.variance, 2)
            gates.RX(qubit, x_angle)
            gates.RZ(qubit, z_angle)
        return qubit


class SystematicUnitaryError(QError):
    '''Simulates a random unitary error that is the same for each qubit'''

    def __init__(self, qchannel, operator = None, variance = None):
        '''
        Instantiate the systematic unitary error class

        :param QChannel qchannel: parent quantum channel
        :param np.array operator:
        :param float variance:
        '''
        QError.__init__(self, qchannel)
        assert operator is not None or variance is not None, \
            "Provide either a random operator or a variance value to use for sampling."

        if operator is not None:
            self.operator = operator
        elif variance is not None:
            x_angle, z_angle = np.random.normal(0, variance, 2)
            Rx = np.cos(x_angle / 2.0) * gates._I - 1j * np.sin(x_angle / 2.0) * gates._X
            Rz = np.cos(z_angle / 2.0) * gates._I - 1j * np.sin(z_angle / 2.0) * gates._Z
            self.operator = np.dot(Rz, Rx)

    def apply(self, qubit):
        '''
        Simulates the application of the unitary error

        :param Qubit qubit: qubit from quantum channel
        :return: rotated qubit
        '''
        if qubit is not None:
            qubit.qsystem.apply(self.operator)
        return qubit



# SQUANCH: Simulator for QUantum Networks and CHannels
# Copyright (c) 2018 AT&T, Ben Bartlett

# Load all modules
from squanch.agent import *
from squanch.channels import *
from squanch.errors import *
from squanch.gates import *
from squanch.linalg import *
from squanch.qstream import *
from squanch.qubit import *
from squanch.simulate import *
